/* blink
 * 
 * A simple driver to blink the tri-color LED on the iMote2.
 *
 * Author: Karl Koscher
 * Version: 0.1
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <asm/hardware.h>
#include <asm-arm/arch-pxa/pxa-regs.h> // This include file lets us access
									   // memory-mapped I/O registers

#include "blink.h"

// These macros include some documentation in the final module file,
// such as a description of the module, the author, and version.
// These macros are optional, but you're encouraged to use them

MODULE_AUTHOR("Karl Koscher");
MODULE_DESCRIPTION("A simple driver to blink the tri-color LED on the iMote2.");
MODULE_VERSION("0.1");

// This macro specifies which license this module is released under.
// This macro is optional, but if you do not specify "GPL", "GPL v2",
// "GPL and additional rights", "Dual BSD/GPL", or "Dual MPL/GPL",
// the kernel will complain that it has been tainted with non-free code.
// Additionally, you may not be able to use certain kernel functions if
// your module is not licensed under the GPL.

// MODULE_LICENSE("GPL");

#define OIER_E4	(1<<4)
#define RED		(1 << 7)

dev_t devId; // Contains the major and minor device numbers
struct cdev *cdev; // A kernel character device struct

int blink_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
static void __exit unload_function(void);

// This struct contains pointers to all supported file operations
struct file_operations blink_fops = {
	.owner = THIS_MODULE,
	.ioctl = blink_ioctl,
};

int state = 0; // State of the LED (on or off)
int delay = 16384; // Period is 1/16384 seconds * delay

// Allow setting the delay as an initialization parameter
module_param(delay, int, 0);

int blink_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg) {
	switch (cmd) {
		case BLINK_SET_RATE:
			OSMR4 = arg; // Update the match register
			OSCR4 = 0; // Reset the counter
			break;

		default:
			return -ENOTTY; // It seems like this should be -EINVAL, but 
							// apparently -ENOTTY also means "so such ioctl"
	}

	return 0;
}

irqreturn_t blink_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	// Check to see if this interrupt is for us
	if (!(OSSR & OIER_E4))
		return IRQ_RETVAL(IRQ_NONE);

	OSSR |= OIER_E4; // Acknowledge this interrupt

	state = !state;

	if (state)
		GPSR3 = RED;
	else
		GPCR3 = RED;

	return IRQ_RETVAL(IRQ_HANDLED);
}

static int __init init_function(void)
{
	int result;

	// Allocate a major device number for this driver
	result = alloc_chrdev_region(&devId, 0, 1, "blink");
	if (result < 0)
		return result;

	// Allocate a character device
	cdev = cdev_alloc();
	if (!cdev) {
		unregister_chrdev_region(devId, 1);
		return -EAGAIN;
	}

	// Set the owner and file operations of this new character device
	cdev->owner = THIS_MODULE;
	cdev->ops = &blink_fops;

	// Register the character device
	// Note at this point the device is live!
	result = cdev_add(cdev, devId, 1);
	if (result < 0) {
		unload_function(); // Tear down everything
		return result;
	}

	result = request_irq(IRQ_OST_4_11, blink_irq_handler, 0, "blink", cdev);
	if (result < 0) {
		unload_function();
		return result;
	}

	// At this point, everything should succeed, so initialize the hardware
	OMCR4 = 0xC9; // Match against channel 4, periodic timer, reset on match,
				  // period is 1 microsecond.
	OSMR4 = delay;
	OIER |= OIER_E4; // Enable interrupts for channel 4
	OSCR4 = 0; // Start the counter

	return 0; // SUCCESS
}

static void __exit unload_function(void)
{
	// Turn off interrupts
	OIER &= ~OIER_E4;
	free_irq(IRQ_OST_4_11, cdev);

	// Free the character device
	cdev_del(cdev);
	// Unregister the major device number
	unregister_chrdev_region(devId, 1);
}

// These two macros let the compiler and kernel know which functions 
// should be called when loading and unloading the kernel.

module_init(init_function);
module_exit(unload_function);

